package Puma::Parser;
use ibase Puma::Core;

=copyright

	Puma - Perl Universal Markup
	Copyright (c) 2001, Ingmar Ellenberger.  All rights reserved.
	Software is distributed AS IS.  No warranties, expressed or implied.

	This license is temporary, pending the selection of an appropriate
	open-source license.

=description

	Parser.pm - The base Puma tag parser

	The Parser module has been written to be flexible and extensible so that it
	can gerve a more generic need rather than any specific one.  For specific
	implementations of the parser, see the modules in the Puma/Parser
	directory.

	Future modifications will likely see this class becoming more generic.

=cut

use Puma::RegEx qw(getRegEx);  # to retrieve the more complex regular expressions
use Puma::Util qq(readfile);   # to read the file in a single chunk

use strict;

# --- standard constructor ---
sub construct {
	my ($self, %args) = @_;

	# --- get the rules for extracting valid labels ---
	$self->stripArgs(\%args, qw(LabelRules Handlers IncludeTag Text FileName));

	return $self->SUPER::construct(%args);
}

=method

	parse - Core parsing method

	This method can be called on any of the following methods:

		parse('text to be parsed')
	or	parse(Text => 'text to be parsed')
	or	parse(FileName => 'file-to-be.parsed')

	It returns a data structure similar to the one created by XML::Parser:

		tagstruct = [ [
			label,
			{
				param => [ value, ... ],
				...
			},
			text | [ tagstruct ]
		] ... ];

=cut

sub parse {
	my $self = shift;
 
	# --- process incoming parameters ---
	unshift(@_, 'Text') if @_ == 1; my %args = @_;  # presume 'Text' if only 1 arg
	$self->stripArgs(\%args, qw(LabelRules Handlers IncludeTag Text FileName));

	# --- load the text to be parsed ---
	delete $self->{text};
	$self->{text} = $self->{Text} if $self->{Text};
	if ($self->{FileName}) {
		$self->{text} = readfile($self->{FileName});
		print "error reading file '$self->{FileName}'\n" unless $self->{text};
	}
	delete @{$self}{qw(Text FileName)};
	return undef unless $self->{text};

	# --- build the root tag ---
	my $tag = {
		label  => ':parser',
		type  => 'indefinite',
		params => {},
	};

	# --- grab the regular expression for finding the tags ---
	# --- store rules in $self so we only have to grab them once ---
	$self->{_rules} = getRegEx('TagParser', $self->{LabelRules});

	# --- recurse! ---
	$self->{_depth} = 0;
	my $retval = $self->_buildTree($tag);

	delete $self->{_depth};
	return $retval;
}

=method

	parseFile - Variant parser for files.

	This is simply a wrapper around the parse() method.

		parseFile('file-to-be.parsed')

	is the same as

		parse(FileName => 'file-to-be.parsed')

=cut

sub parseFile {
	my ($self, $filename) = @_;
	return $self->parse(FileName => $filename);
}

=method

	buildTag - a helper function to build a tag structure

	This method takes two or more parameters and returns a tag structure in the
   format specified above.

	The two parameters are the tag label and a hasshref of tag paramters.  All
   following fields are text or references to child tags.

=cut

sub buildTag {
	my ($self, $label, $params, @children) = @_;
	return [ $label, $params || undef, @children ];
}

=Private Methods ============================================================
=method

	_buildTree - Build the tag tree (structure)

	This method calls itself recursively to build a tree of tags.

=cut

sub _buildTree {
	my ($self, $tag) = @_;

	# --- don't recurse more than 40 levels deep ---
	return if $self->{_depth} > 40;
	$self->{_depth}++;

	# --- use precompiled rules to search through text and build tag tree ---
	my @children;
	TAG: while ($self->{text} =~ m/$self->{_rules}/gco) { # '/o' to compile regex once
		push @children, $1 if $1;  # push text as a child

		my $child = $self->_splitTag($2);

		# --- check handlers for matching tag label ---
		foreach my $key (keys %{$self->{Handlers}}) {
			if ($child->{label} =~ /$key/) {
				my @handlertags = &{$self->{Handlers}->{$key}}($self, $child);
				push @children, @handlertags if @handlertags;
				next TAG;
			}
		}

		if ($child->{type} =~ /code|eval|selfending/) {
			push @children, $self->buildTag($child->{label}, $child->{params});
		} elsif ($child->{type} eq 'beginning') {
			push @children, $self->_buildTree($child);
		} elsif ($child->{type} eq 'ending') {
			last if $child->{label} eq $tag->{label};
		}
		# TODO: else generate mismatched tag error
	}

	# --- grab text between last tag and EOF ---
	if ($self->{_depth} == 1) {
		$self->{text} =~ m/\G(.*)/gs;
		push @children, $1 if $1;  # push text as a child
	}

	# --- return parsed tree ---
	$self->{_depth}--;
	return $self->buildTag($tag->{label}, $tag->{params}, @children);
}

=method

	_splitTag - Breaks the tag into a label and key/value pairs

=cut

sub _splitTag {
	my ($self, $text) = @_;

	my $tag;
	# --- determine the tag type ---
	SWITCH: {
		# --- in-line code ---
		($text =~ /^<:\s/) && do {
			$tag->{type} = 'code';
			$tag->{label} = ':';
			last;
		};
		# --- in-line code returning a display value ---
		($text =~ /^<:=\s/) && do {
			$tag->{type} = 'eval';
			$tag->{label} = ':=';
			last;
		};
		# --- self-ending tag ---
		($text =~ m|/>$|) && do { $tag->{type} = 'selfending'; last; };
		# --- ending tag ---
		($text =~ m|^</|) && do { $tag->{type} = 'ending'; last; };
		# --- beginning tag ---
		$tag->{type} = 'beginning';
	}

	# --- process code and eval tags ---
	if ($tag->{type} =~ /code|eval/) {
		# TODO: detect and throw error when tag not closed properly.
		$tag->{params}->{$tag->{type}} = ($text =~ m|<:=?\s(.*?)/>$|)[0];
		return $tag;
	}


	# === process all other (not code or eval) tags ===
	# --- get the label ---
	$text =~ m|</?([^>\s/]+)\s*(?:/?>)?|g;
	$tag->{label} = $1 if $1;

	my $rules = getRegEx('TagSplitter');
	my $lastparam = 0;
	while (!$lastparam) {
		if ($text =~ m/$rules/gco) {

			# --- found a paired value ---
			my ($key, $value) = ($1, $+);
			$value =~ s/\\([`'"])/$1/gs;  # normalize escapes
			$tag->{params}->{$key} = $value;

		} elsif ($text =~ m/\G([\w.-]+)\s*/gc) {

			# --- found a key with no value ---
			$tag->{params}->{$1} = '';

		} else {

			# --- we've hit the last parameter ---
			$lastparam = 1;
		}
	}

	return $tag;
}

1;
